/****************************************************************************
*** .INI file module of Remote Agent PC 
*** Copyright 1999 Demmel Hard & Soft

The .INI file has the same format as Windows .INI files. The maximum
line length is limited to MAX_INI_LINE_LEN - 1 (inifiled.h). Section names
are enclosed in brackets. Comments start with a semicolon and extend to the
end of the line. End-of-line markers can be CR, CR/LF or LF.

HISTORY:
WHEN			WHO	WHAT
----------------------------------------------------------------------------
1998-10-28	Jerry	Creation
1999-03-29	dh - Adapted for SERIALIO project

****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "inifile.h"

// Size of read buffer for read_line() function
#define READBUF_SIZE		1024

// Maximum length of a DOS path
#define MAXPATH	80

// Possible states for end of line
enum eol_state
{
	eol_not,		// Currently not at an end of line
	eol_cr,		// Just before a \r
	eol_lf,		// Just before a \n
	eol_past		// Just past an end of line
};

// Name of .INI file (without path)
static char ini_fname[MAXPATH + 1];

// Read buffer for read_line function
static char ini_readline_buf[READBUF_SIZE];

// File handle of .INI file
static FILE *ini_fh;

// TRUE if end of .INI file is reached
static int ini_eof;

// TRUE if .INI file exhausted (all read there is to read)
static int ini_exhausted;

// Pointer to current position in ini_readline_buf[]
static char *ini_currpos;

// Pointer after the last valid character in ini_readline_buf (if the end
// of file is reached, the buffer will not be filled completely)
static char *ini_eobuffer;

// TRUE if ini_currpos points to an end of line (CRLF) of a line which was
// already returned and should therefore be skipped
static enum eol_state ini_skip_eol;

// Number of current line in .INI file
static int ini_linenr = 0;

// Scan forward for char c in s for a maximum of cnt bytes.
// Return zero if char c is not found, else return byte position (1-based).
static int byte_pos(int c, char *s, unsigned cnt)
{
	word i;

	for (i = 0; i < cnt; i++)
	{
		if (s[i] == c)
			return i + 1;
	}

	return 0;
}

// Scan forward for char c in string s and return zero if char c is not found,
// else return byte position (1-based).
static int str_pos(int c, char *s)
{
	word i, cnt = strlen(s);

	for (i = 0; i < cnt; i++)
	{
		if (s[i] == c)
			return i + 1;
	}

	return 0;
}

// Return one line from the specified file
// s ... string the line is to be returned in
// slen ... length of string including \0
// Returns FALSE if end of file reached, TRUE otherwise
static int raw_readline(char *s, unsigned slen)
{
	int eol_pos = 0;	// 1-based position of end of line or 0 if eol not found

	// If string has zero length, return immediately
	if (!slen)
		return 0;

	// Reserve one character for string terminator
	--slen;

	// Read until end of file or line is reached or passed string is full
	while (!ini_eof && !eol_pos && slen)
	{
		// If current position is at the end of the read buffer, read in
		// new block
		if (ini_currpos == ini_eobuffer)
		{
			if (ini_exhausted)
				ini_eof = TRUE;
			else
			{
				int bytes_read = fread(ini_readline_buf, 1, READBUF_SIZE, ini_fh);

				// Check if end of file is reached
				ini_exhausted = bytes_read < READBUF_SIZE;
				ini_eof = !bytes_read;

				// Set new end of buffer and current pointer
				ini_eobuffer = ini_readline_buf + bytes_read;
				ini_currpos = ini_readline_buf;
			}
		}

		// Remove end-of-line character(s) if there are any
		if (ini_currpos != ini_eobuffer &&
			(ini_skip_eol == eol_cr || ini_skip_eol == eol_lf))
		{
			// If we are here, we must have at least one character and it must
			// be a \r or a \n; skip past it
			++ini_currpos;

			// If we just skipped over a \r, next character must be a \n,
			// otherwise we skipped a \n and are finished
			if (ini_skip_eol == eol_cr)
				ini_skip_eol = eol_lf;
			else
				ini_skip_eol = eol_past;

			// If we still have a character available and have to skip over
			// the \n, Just Do It (TM)
			if (ini_skip_eol == eol_lf && ini_currpos != ini_eobuffer)
			{
				if (*ini_currpos == '\n')
					++ini_currpos;

				ini_skip_eol = eol_past;
			}
		}

		// If the current position is at the end of the buffer, the skipping
		// over the end of line used up all available characters in the buffer
		// so we must go back to read in a new block
		if (ini_currpos != ini_eobuffer)
		{
			// Number of unprocessed characters in buffer
			unsigned buf_len = ini_eobuffer - ini_currpos;

			// Max number of characters to be checked
			unsigned max_len = min(buf_len, slen);

			// Number of characters to copy
			unsigned nr2copy = max_len;

			// If we are currently just past an eol, increment line counter
			if (ini_skip_eol == eol_past)
			{
				++ini_linenr;
				ini_skip_eol = eol_not;
			}

			// Look for carriage return or line feed
			if ((eol_pos = byte_pos('\r', ini_currpos, max_len)) != 0 ||
				(eol_pos = byte_pos('\n', (byte *) ini_currpos, max_len)) != 0)
			{
				// End of line character could be found
				// Copy only up to eol
				nr2copy = eol_pos - 1;

				// Set flag that we encountered an eol
				ini_skip_eol = ini_currpos[eol_pos - 1] == '\r' ? eol_cr : eol_lf;
			}

			// If eol is the first character, we don't need to copy anything
			if (nr2copy)
			{
				// Copy back buffer until eol or end of buffer
				memmove(s, ini_currpos, nr2copy);

				// Set string pointer past end of already filled string
				s += nr2copy;

				// Set current position past end of processed part of buffer
				ini_currpos += nr2copy;

				// Adjust number of available characters in string
				slen -= nr2copy;
			}
		}
	}

	// Terminate string
	*s = '\0';

	// Return FALSE if end of file reached
	return !ini_eof;
}


// Return next non-empty (ignoring whitespace), non-comment-only line from
// the .INI file. Comments and leading and trailing whitespace are removed.
// Same parameters as raw_readline
// s ... string the line is to be returned in
// slen ... length of string including \0
// Returns pointer to real begin of line (ignoring whitespace), or 0 if there
// are no more lines
static char *cooked_readline(char *s, unsigned slen)
{
	while (raw_readline(s, slen))
	{
		// If line is not empty
		if (*s)
		{
			// Remove comment if there is one
			word comment_pos = str_pos(';', s);		// !!

			if (comment_pos)
			{
				// If the semicolon is preceeded by a backslash, it is not a
				// start-of-comment
				if (comment_pos > 1 & s[comment_pos - 2] == '\\')
					// Remove backslash (shift string to the left)
					memmove(s + comment_pos - 2,	s + comment_pos - 1, 
						strlen(s + comment_pos - 1) + 1);
				else
					// Remove comment
					s[comment_pos - 1] = '\0';
			}

			// Remove leading whitespace (we cannot use str_[lt]marg because they
			// only handle spaces and not tabs)
			while (*s == ' ' || *s == '\t')
				++s;

			// If there are characters left in the string
			if (*s)
			{
				// Remove trailing spaces
				char *t = s + strlen(s) - 1;

				while (t != s && (*t == ' ' || *t == '\t'))
					*t-- = '\0';

				// If string is still not empty, return it otherwise read next line
				if (*s)
					return s;
			}
		}
	}

	// Reached end of file
	return 0;
}


// Return next line after "cooking" (see cooked_readline) but stop at
// a new section
// s ... string the line is to be returned in
// slen ... length of string including \0
// Returns pointer to real begin of line (ignoring whitespace), or 0 if there
// are no more lines
char *ini_section_readline(char *s, unsigned slen)
{
	char *line = cooked_readline(s, slen);

	// If we encountered a new section beginning, stop
	if (line[0] == '[' && line[strlen(line) - 1] == ']')
		return 0;

	return line;
}


// Go to the begin of the .INI file
static void ini_rewind(void)
{
	if (ini_fh != NULL)
	{
		// Set file pointer to begin of file
		rewind(ini_fh);

		// Initialize some variables used by readline()
		ini_exhausted = FALSE;
		ini_eof = FALSE;
		ini_currpos = ini_eobuffer = ini_readline_buf;
		ini_skip_eol = eol_not;
		ini_linenr = 1;
	}
}


// Open the .INI file looking for it in the current directory and the
// program directory
// Returns TRUE if open was successful, FALSE otherwise
int ini_open(char *fname)
{
	strcpy(ini_fname, fname);

	// Try to open .INI file read-only in the current directory
	ini_fh = fopen(ini_fname, "r");	

	// If it could not be opened
	if (ini_fh == NULL)
		return FALSE;

	// .INI file opened successfully; initialize some variables
	ini_rewind();

	return TRUE;
}

// Close the .INI file
int ini_close(void)
{
	if (ini_fh != NULL)
	{
		fclose(ini_fh);
		ini_fh = NULL;
		return TRUE;
	}
	else
		return FALSE;
}


// Go to the specified section of the .INI file (subsequent ini_readline()s
// will read the lines past the section header
// section_name ... Name of the section to go to (without the brackets)
// Returns TRUE if section was found, FALSE otherwise
int ini_goto_section(char *section_name)
{
	char linebuff[MAX_INI_LINE_LEN], *line;

	// If .INI file was not opened, return immediately
	if (ini_fh == NULL)
		return FALSE;

	// Go to begin of file
	ini_rewind();

	// Read lines until eof reached or section found
	while ((line = cooked_readline(linebuff, MAX_INI_LINE_LEN)) != NULL)
	{
		// Check if section header found
		if (line[0] == '[')
		{
			// Locate end of section name
			word end_pos = str_pos(']', line);

			// If specified section found, return
			if (end_pos > 2 && !strnicmp(line + 1, section_name, end_pos - 2))
				return TRUE;
		}
	}
	// Section not found
	return FALSE;
}


// Return the current line number of the .INI file
int ini_curr_line(void)
{
	return ini_linenr;
}

// Return the value of the specified variable in the specified section as a
// string
// section ... name of section in which variable is to be looked for
// varname ... name of variable whose value is to be returned
// s ... pointer to line buffer
// slen ... real length of s
// Returns 0 if section or variable not found
char *ini_get_varstr(char *section, char *varname, char *s, unsigned slen)
{
	char *line;

	// If specified section not found, return immediately
	if (!ini_goto_section(section))
		return 0;

	// Read lines of section until variable found or end of section reached
	while ((line = ini_section_readline(s, slen)) != NULL)
	{
		char *vmatch = varname;

		// Check if variable names match (up to whitespace or equal sign)
		while (*vmatch && *line && tolower(*vmatch) == tolower(*line))
		{
			++vmatch;
			++line;
		}

		// If variable names match (end of varname reached and variable name
		// in file followed by whitespace or equal sign)
		if (!*vmatch && (*line == ' ' || *line == '\t' || *line == '='))
		{
			// Get position of equal sign
			word equ_pos = str_pos('=', line);

			if (equ_pos)
			{
				// Skip over equal sign
				line += equ_pos;

				// Skip whitespace before value
				while (*line == ' ' || *line == '\t')
					++line;

				// Return pointer to value
				return line;
			}
		}
	}

	// Variable not found in section
	return 0;
}


// Remove all whitespace from given string
// Returns s.
char *ini_remove_whitespace(char *s)
{
	word len = strlen(s);
	char *curr_char = s;

	// Until end of string is reached
	while (*curr_char)
	{
		// If current character is whitespace, move rest of string over it
		if (*curr_char == ' ' || *curr_char == '\t')
			memmove(curr_char, curr_char + 1, len);
		else
			++curr_char;

		--len;
	}

	return s;
}
