/****************************************************************************
*** SERIALIO module of Remote Agent PC 
*** Copyright 1999 Demmel Hard & Soft

HISTORY:
WHEN			WHO	WHAT
----------------------------------------------------------------------------
1999-03-29	dh -	Creation
1999-06-17	dh -	Fixed problem with interrupt in
						sio_OpenDigiboard() (pending interrupt after re-initialize
						before setting the parameters correctly)

****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "global.h"
#include "ports.h"
#include "serialio.h"

// CAUTION: !!! Stack checking must be turned off for interrupt routines !!!
// (MSC only, Borland has it disabled by default.)
#pragma check_stack(off)


typedef struct
{
	word base_addr;
	int interrupt_no;
	int digiboard_no;
	int initialized;
	queue_t rx_queue;
	int queue_overrun;
} com_t;

typedef struct
{
	int first_port_no;
	int first_port_addr;
	word status_reg_addr;
	int no_of_ports;
	int interrupt_no;
} digi_t;

typedef struct
{
	ISRPtr old_int_vec;
	word old_mask;
	word old_mask2;
} sio_OldInts_t;

#define SIO_MAX_PORTS		16
#define SIO_MAX_INTS			16
#define SIO_MAX_DIGIBOARDS	2		// CAUTION: Do not modify

// all COM ports are described in this structure
static com_t sio_Ports[SIO_MAX_PORTS] = { 0 };

// all Digiboards are described in this structure
static digi_t sio_Digiboards[SIO_MAX_DIGIBOARDS] = { 0 };

static sio_OldInts_t sio_OldInts[SIO_MAX_INTS] = { NULL, 0 };
static int sio_ValidInts[] = { 3, 4, 5, 10, 11, 12, 15, 0 };	// for RAPC
// static int sio_ValidInts[] = { 3, 4, 5, 9, 10, 11, 12, 15, 0 };	// others

#define sio_BAddr(PortNo)	sio_Ports[PortNo].base_addr

static void sio_CommonHandler(int portno)
{
	static word status, data, base_addr;
	static queue_t *q;

#if 0			// for "debugging" only
	word far *p;
	byte id;

	p = MK_FP(0xb800, 0);
	*p = 0x700 + 'I';
#endif

	base_addr = sio_BAddr(portno);

   while (TRUE)
   {
      switch (IN(INTID(base_addr)) & 0x0f)
      {
         case IID_REC_DATA:	// received data
         case IID_TIMEOUT:		// got timeout
         case IID_LINE_STAT:	// line status changed
            do
				{
					// Caution: Status MUST be read before data, otherwise a 
					// pending break may be lost !

					status = IN(LINSTAT(base_addr));

					if (status & DR)
	               data = IN(RXREG(base_addr)) | ((status & HARD_ERROR) << 8);
					else
						data = (status & HARD_ERROR) << 8;

					status = IN(LINSTAT(base_addr));

					// write data to buffer (if there is enough space)
					q = &sio_Ports[portno].rx_queue;

					DISABLE();
					if (queue_Exists(q) && queue_IsSpaceInQueue(q))
					{
						queue_FillQueue(q, data);
					}
					else if (queue_Exists(q))
						sio_Ports[portno].queue_overrun = TRUE;

					ENABLE();

            } while (status & DR);

            break;

         case IID_NO_INTR:		// no interrupt
            return;				// way out of the "while (TRUE)" loop

         case IID_TX_EMPTY:	// transmit shift register empty
            break;				// do nothing - we don't use TX interrupts 

         case IID_MODEM_STAT:
            data = (word) IN(MODSTAT(base_addr)) | (word) ((word) STAT_CHNG << 8);

				// write data to buffer (if there is enough space)
				q = &sio_Ports[portno].rx_queue;

				DISABLE();
				if (queue_Exists(q) && queue_IsSpaceInQueue(q))
				{
					queue_FillQueue(q, data);
				}
				else if (queue_Exists(q))
					sio_Ports[portno].queue_overrun = TRUE;

				ENABLE();

            break;

         default:					// unexpected interrupt
            return;
      }
   }
}

static void interrupt far sio_COMInt1(void)
{
	ENABLE();
	sio_CommonHandler(COM1);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Ports[COM1].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static void interrupt far sio_COMInt2(void)
{
	ENABLE();
	sio_CommonHandler(COM2);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Ports[COM2].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static void interrupt far sio_COMInt3(void)
{
	ENABLE();
	sio_CommonHandler(COM3);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Ports[COM3].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static void interrupt far sio_COMInt4(void)
{
	ENABLE();
	sio_CommonHandler(COM4);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Ports[COM4].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static void interrupt far sio_DigiboardInt0(void)
{
   byte status;

	ENABLE();

   while ((status = IN(sio_Digiboards[0].status_reg_addr)) != 0xff)
      sio_CommonHandler(sio_Digiboards[0].first_port_no + status);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Ports[COM3].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static void interrupt far sio_DigiboardInt1(void)
{
	byte status;

	ENABLE();

   while ((status = IN(sio_Digiboards[1].status_reg_addr)) != 0xFF)
      sio_CommonHandler(sio_Digiboards[1].first_port_no + status);

	DISABLE();
	OUT(PIC_ADDR, PIC_EOI);	   				// pic end of interrupt
	if (sio_Digiboards[0].interrupt_no >= 8)
		OUT(PIC2_ADDR, PIC_EOI);				// pic 2 end of interrupt
}

static boolean sio_IsIntValid(int int_no)
{
	int int_valid = FALSE;
	int i;

	for (i = 0; sio_ValidInts[i] != 0; i++)
	{
		if (int_no == sio_ValidInts[i])
		{
			int_valid = TRUE;
			break;
		}
	}
	
	if (!int_valid)
		return FALSE;

	return TRUE;
}

static int sio_IsPortNoValid(int portno)
{
	if (portno < 0 || portno >= SIO_MAX_PORTS)
		return FALSE;

	return TRUE;
}

int sio_IsPortOpen(int portno, int *errno)
{
	*errno = SIO_NO_ERR;

	if (!sio_IsPortNoValid(portno))
	{
		*errno = SIO_INVAL_PORTNO;
		return FALSE;
	}

	if (sio_Ports[portno].base_addr == 0)
		return FALSE;

	return TRUE;
}

static int sio_IsDigiboardNoValid(int digiboard_no)
{
	if (digiboard_no < 0 || digiboard_no > SIO_MAX_DIGIBOARDS)
		return FALSE;

	return TRUE;
}

int sio_IsDigiboardOpen(int digiboard_no, int *errno)
{
	*errno = SIO_NO_ERR;

	if (!sio_IsDigiboardNoValid(digiboard_no))
	{
		*errno = SIO_INVAL_DB_NO;
		return FALSE;
	}

	if (sio_Digiboards[digiboard_no].first_port_addr == 0)
		return FALSE;

	return TRUE;
}

static int sio_SetCOMParm(int portno, dword baudrate, int parity,
					int stop, int databits)
{
	// set communication parameters

   unsigned char parmbyte;
   unsigned int divisor;
   unsigned char lsb, msb;
	int err;
  
	if (!sio_IsPortOpen(portno, &err))
	{
		if (err)
			return err;

		return SIO_PORT_NOPEN;
	}

   // First, calculate the baud rate driver 
   divisor = (unsigned) (115200L / (long) baudrate);
   msb = (unsigned char) (divisor >> 8);
   lsb = (unsigned char) ((divisor << (unsigned) 8) >> 8);

	// Disable int's 
   OUT(sio_INT_EN(portno), 0x0);
  
   // Enable access to the divisor latches by setting
   // the divisor latch access bit in the line control register 
   OUT(sio_LINCONT(portno), DLAB);
  
   // Set the least significant byte 
   OUT(sio_DIVLL(portno), lsb);
  
   // Set the most significant (high) byte 
   OUT(sio_DIVLH(portno), msb);
  
   // Start to calculate the parameter byte 
  
   // First, the word length. This occupies bit 0 and 1 of the parameter
   // byte from 0 0 for 5 bit words to 1 1 for 8 bit words 

   parmbyte = (unsigned char) ((databits - 5) & (WLENGTH_L | WLENGTH_M));
  
   // Next, the stop bit
   parmbyte |= (stop > 1) ? STOPBIT2 : 0;
  
   // Now, the parity 

   if (parity != NONE)
      parmbyte |= PENABLE;								// Parity enable 
  
   if (parity == EVEN)
      parmbyte |= PSELECT;								// even
  
   OUT(sio_LINCONT(portno), parmbyte);				// output it to UART

	// we have a FIFO on every serial port, so don't check existance, 
	// simple enable it with default value of 8 bytes trigger
	OUT(sio_FIFOCONT(portno), F_RX_8 | F_RX_RES | F_TX_RES | F_EN);

	return SIO_NO_ERR;
}
  
static int sio_HookInterrupt(int int_no, ISRPtr handler)
{
	if (!sio_IsIntValid(int_no))
		return SIO_INVAL_INT;

	if (sio_OldInts[int_no].old_int_vec != NULL)
		return SIO_ALR_HOOKED;

	DISABLE();		// disable interrupts

	// get old serial interrupt 
	if (int_no < 8)
		sio_OldInts[int_no].old_int_vec = GETVECT(int_no + 0x08);
	else 
		sio_OldInts[int_no].old_int_vec = GETVECT(int_no - 8 + 0x70);

	// set new interrupt
	SETVECT((int_no < 8) ? (int_no + 0x08) : (int_no - 8 + 0x70), handler);

	// enable interrupt controller mask(s):

	if (int_no < 8)
	{
		OUT(PIC_MASK, (sio_OldInts[int_no].old_mask = IN(PIC_MASK)) & (~(1 << int_no)));
		sio_OldInts[int_no].old_mask2 = 0xffff;
	}
	else
	{
		OUT(PIC_MASK, (sio_OldInts[int_no].old_mask = IN(PIC_MASK)) & (~(1 << 2)));
		OUT(PIC2_MASK, (sio_OldInts[int_no].old_mask2 = IN(PIC2_MASK)) & (~(1 << (int_no - 8))));
	}

	ENABLE();		// re-enable interrupts

	return SIO_NO_ERR;
}

static int sio_UnHookInterrupt(int int_no)
{
	if (!sio_IsIntValid(int_no))
		return SIO_INVAL_INT;
	
	if (sio_OldInts[int_no].old_int_vec == NULL)
		return SIO_ALR_UNHOOKED;

	DISABLE();

	// disable interrupt controller mask(s):
	OUT(PIC_MASK, sio_OldInts[int_no].old_mask);
	sio_OldInts[int_no].old_mask = 0x0;

	if (int_no >= 8)
		OUT(PIC2_MASK, sio_OldInts[int_no].old_mask2);
	
	sio_OldInts[int_no].old_mask2 = 0x0;

	if (int_no < 8)
		SETVECT(int_no + 0x08, sio_OldInts[int_no].old_int_vec);
	else
		SETVECT(int_no - 8 + 0x70, sio_OldInts[int_no].old_int_vec);

	sio_OldInts[int_no].old_int_vec = NULL;

	ENABLE();
	return SIO_NO_ERR;
}

static int sio_PortExists(word port_addr)
{
	OUT(SCRATCH(port_addr), 0x55);
	if (IN(SCRATCH(port_addr)) != 0x55)
		return FALSE;

	OUT(SCRATCH(port_addr), 0xAA);
	if (IN(SCRATCH(port_addr)) != 0xAA)
		return FALSE;

	return TRUE;
}

int sio_OpenDigiboard(int digiboard_no, int first_port_no, int no_of_ports,
		word first_port_addr, word status_reg_addr, int interrupt_no)
{
	int i, err;
	word addr = first_port_addr;
	ISRPtr isrp = NULL;

	if (!sio_IsDigiboardNoValid(digiboard_no))
		return SIO_INVAL_DB_NO;

	if (sio_IsDigiboardOpen(digiboard_no, &err))
	{
		if (err)
			return err;
		else
			return SIO_DIGI_ALR_OPEN;
	}

	if (first_port_no + no_of_ports > SIO_MAX_PORTS ||
		no_of_ports <= 0 || first_port_no <= COM2)
		return SIO_INVAL_PORTNO;

	if (first_port_addr == 0)
		return SIO_INVAL_ADDR;

	switch (digiboard_no)
	{
		case 0:
			isrp = sio_DigiboardInt0;
			break;
		case 1:
			isrp = sio_DigiboardInt1;
			break;
		default:
			return SIO_INVAL_DB_NO;
	}

	sio_Digiboards[digiboard_no].first_port_no = first_port_no;
	sio_Digiboards[digiboard_no].first_port_addr = first_port_addr;
	sio_Digiboards[digiboard_no].status_reg_addr = status_reg_addr;
	sio_Digiboards[digiboard_no].no_of_ports = no_of_ports;
	sio_Digiboards[digiboard_no].interrupt_no = interrupt_no;

	for (i = first_port_no; i < first_port_no + no_of_ports; i++)
	{
		if (sio_IsPortOpen(i, &err))
		{
			sio_CloseDigiboard(digiboard_no);

			if (err)
				return err;

			return SIO_PORT_USE;
		}

		if (!sio_PortExists(addr))
		{
			sio_CloseDigiboard(digiboard_no);
			return SIO_NO_PORT;
		}

		sio_Ports[i].base_addr = addr;
		sio_Ports[i].interrupt_no = interrupt_no;
		sio_Ports[i].digiboard_no = digiboard_no + 1;	// we use idx + 1
		sio_Ports[i].queue_overrun = FALSE;

		addr += 8;
	}

	if ((err = sio_HookInterrupt(interrupt_no, isrp)) != SIO_NO_ERR)
	{
		sio_CloseDigiboard(digiboard_no);
		return err;
	}

	return SIO_NO_ERR;
}

int sio_CloseDigiboard(int digiboard_no)
{
	int int_no, i, err = 0;

	if (!sio_IsDigiboardOpen(digiboard_no, &err))
	{
		if (err)
			return err;

		return SIO_DIGI_NOPEN;
	}

	int_no = sio_Digiboards[digiboard_no].interrupt_no;

	for (i = sio_Digiboards[digiboard_no].first_port_no;
		i < sio_Digiboards[digiboard_no].first_port_no +
			sio_Digiboards[digiboard_no].no_of_ports; i++)
	{
		sio_Ports[i].base_addr = 0;
		sio_Ports[i].interrupt_no = 0;
		sio_Ports[i].digiboard_no = 0;
		sio_Ports[i].initialized = FALSE;
		sio_Ports[i].queue_overrun = FALSE;

		if (sio_Ports[i].rx_queue.data != NULL)
		{
			if (err == 0)
				err = queue_DeInit(&sio_Ports[i].rx_queue);
			else
				queue_DeInit(&sio_Ports[i].rx_queue);
		}
	}

	sio_Digiboards[digiboard_no].first_port_no = 0;
	sio_Digiboards[digiboard_no].first_port_addr = 0;
	sio_Digiboards[digiboard_no].status_reg_addr = 0;
	sio_Digiboards[digiboard_no].no_of_ports = 0;
	sio_Digiboards[digiboard_no].interrupt_no = 0;

	if (err)
	{
		sio_UnHookInterrupt(int_no);
		return err;
	}

	return sio_UnHookInterrupt(int_no);
}

int sio_GetDigiboardInfo(int digiboard_no, int *first_port, int *no_of_ports)
{
	int err;

	if (!sio_IsDigiboardOpen(digiboard_no, &err))
	{
		if (err)
			return err;

		return SIO_DIGI_NOPEN;
	}

	*first_port = sio_Digiboards[digiboard_no].first_port_no;
	*no_of_ports = sio_Digiboards[digiboard_no].no_of_ports;

	return SIO_NO_ERR;
}

int sio_OpenCOMPort(int port_no, word port_addr, int interrupt_no)
{
	int err = 0;
	ISRPtr isrp = NULL;

	if (port_no > COM4 || port_no < COM1)
		return SIO_INVAL_PORTNO;

	if (sio_IsPortOpen(port_no, &err))
	{
		if (err)
			return err;

		return SIO_PORT_USE;
	}

	if (port_addr == 0)
		return SIO_INVAL_ADDR;

	if (!sio_PortExists(port_addr))
		return SIO_NO_PORT;

	switch (port_no)
	{
		case COM1:
			isrp = sio_COMInt1;
			break;
		case COM2:
			isrp = sio_COMInt2;
			break;
		case COM3:
			isrp = sio_COMInt3;
			break;
		case COM4:
			isrp = sio_COMInt4;
			break;
		default:
			return SIO_INVAL_DB_NO;
	}

	sio_Ports[port_no].base_addr = port_addr;
	sio_Ports[port_no].interrupt_no = interrupt_no;
	sio_Ports[port_no].digiboard_no = 0;

	if ((err = sio_HookInterrupt(interrupt_no, isrp)) != SIO_NO_ERR)
	{
		sio_Ports[port_no].base_addr = 0;
		sio_Ports[port_no].interrupt_no = 0;

		return err;
	}

	return SIO_NO_ERR;
}

int sio_CloseCOMPort(int port_no)
{
	int int_no;
	int err = 0;

	if (!sio_IsPortOpen(port_no, &err))
	{
		if (!err)
			err = SIO_PORT_NOPEN;
	}

	if (sio_Ports[port_no].digiboard_no > 0)
	{
		// Port belongs to a digiboard group, close with appropriate function

		return SIO_NOT_STD_COM;
	}

	// disable UART interrupts
   OUT(sio_INT_EN(port_no), 0);

	int_no = sio_Ports[port_no].interrupt_no;

	sio_Ports[port_no].base_addr = 0;
	sio_Ports[port_no].interrupt_no = 0;
	sio_Ports[port_no].digiboard_no = 0;
	sio_Ports[port_no].initialized = FALSE;
	sio_Ports[port_no].queue_overrun = FALSE;
	
	if (sio_Ports[port_no].rx_queue.data != NULL)
	{
		if (err == 0)
			err = queue_DeInit(&sio_Ports[port_no].rx_queue);
		else
			queue_DeInit(&sio_Ports[port_no].rx_queue);
	}

	if (err)
	{
		sio_UnHookInterrupt(int_no);
		return err;
	}

	return sio_UnHookInterrupt(int_no);
}

int sio_CloseAllPorts(void)
{
	int errno = 0, dummy, open, i;

	for (i = 0; i < SIO_MAX_DIGIBOARDS; i++)
	{
		if (errno)
			open = sio_IsDigiboardOpen(i, &dummy);
		else 
			open = sio_IsDigiboardOpen(i, &errno);

		if (open)
		{
			if (errno)
				sio_CloseDigiboard(i);
			else
				errno = sio_CloseDigiboard(i);
		}
	}

	for (i = 0; i < SIO_MAX_PORTS; i++)
	{
		if (errno)
			open = sio_IsPortOpen(i, &dummy);
		else 
			open = sio_IsPortOpen(i, &errno);

		if (open)
		{
			if (errno)
				sio_CloseCOMPort(i);
			else
				errno = sio_CloseCOMPort(i);
		}
	}

	return errno;
}

int sio_InitPort(int portno, dword baudrate, int parity,
					int stop, int databits, word interrupt_mask)
{
	int err;

	if (!sio_IsPortOpen(portno, &err))
	{
		if (err)
			return err;

		return SIO_PORT_NOPEN;
	}

	if ((err = sio_SetCOMParm(portno, baudrate, parity, stop, databits)) != 0)
		return err;

	if (sio_Ports[portno].rx_queue.data == NULL)
	{
		// initialize queue

		if ((err = queue_Init(&sio_Ports[portno].rx_queue, DEF_QUEUE_SIZE)) != 0)
			return err;
	}

   OUT(sio_MODCONT(portno), DTR | RTS | OUT1 | OUT2);
   OUT(sio_INT_EN(portno), interrupt_mask);

	sio_Ports[portno].initialized = TRUE;
	return SIO_NO_ERR;
}

int sio_IsPortInitialized(int portno, int *errno)
{
	*errno = 0;

	if (!sio_IsPortOpen(portno, errno))
		return FALSE;

	return sio_Ports[portno].initialized;
}

int sio_WasQueueOverrun(int portno)
{
	return sio_Ports[portno].queue_overrun;
}

int sio_ResetQueueOverrun(int portno)
{
	int err;

	if (!sio_IsPortInitialized(portno, &err))
	{
		if (err)
			return err;

		return SIO_PORT_NOT_INIT;
	}

	return sio_Ports[portno].queue_overrun = FALSE;
}

int sio_ClearRXQueue(int portno)
{
	int err;

	if (!sio_IsPortInitialized(portno, &err))
	{
		if (err)
			return err;

		return SIO_PORT_NOT_INIT;
	}
	
	DISABLE();
	queue_ResetQueue((&sio_Ports[portno].rx_queue));
	ENABLE();

	return SIO_NO_ERR;
}

void sio_SendChar(int portno, int c)
{
	int addr = sio_LINSTAT(portno);

   while (!(IN(addr) & TSRE))
      ;												// wait for transmit char. ready
  
	addr = sio_TXREG(portno);
   OUT(sio_TXREG(portno), c & 0xff);		// send it 
}

void sio_SendString(int portno, char *s)
{
	while (*s)
		sio_SendChar(portno, *s++);
}

void sio_SendBuffer(int portno, byte *s, int len)
{
	for (; len > 0; len--)
		sio_SendChar(portno, *s++);
}

int sio_GetOne(int portno, word *s)
{
	register int got = 0;
	queue_t *q = &sio_Ports[portno].rx_queue;

	if (queue_IsQueueNotEmpty(q))
	{
		DISABLE();
		queue_ReadQueue(q, *s);
		ENABLE();

		got++;
	}

	return got;
}

int sio_GetBuffer(int portno, word *s, word maxsize)
{
	register int data;
	register int got = 0;
	queue_t *q = &sio_Ports[portno].rx_queue;

	while (maxsize > 0)
	{
		if (queue_IsQueueNotEmpty(q))
		{
			DISABLE();
			queue_ReadQueue(q, data);
			ENABLE();

			*s++ = data;
			got++;
			maxsize--;
		}
		else
			break;
	}

	return got;
}

word sio_GetModemStatus(int portno)
{
   return IN(sio_MODSTAT(portno));
}

void sio_SetModemControl(int portno, word value)
{
	value &= DTR | RTS;
	value |= OUT1 | OUT2;

	OUT(sio_MODCONT(portno), value);
}

void sio_BreakOn(int portno)
{
	int line_reg;

	line_reg = inp(sio_LINCONT(portno));
	outp(sio_LINCONT(portno), (line_reg | BREAKOUT));
}

void sio_BreakOff(int portno)
{
	int line_reg;

	line_reg = inp(sio_LINCONT(portno));
	outp(sio_LINCONT(portno), line_reg & (~BREAKOUT));
}
